# SPDX-FileCopyrightText: 2019 Nicholas H. Tollervey for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""
`adafruit_ble_radio`
================================================================================

Simple byte and string based inter-device communication via BLE.


* Author(s): Nicholas H.Tollervey for Adafruit Industries

**Hardware:**

   Adafruit Feather nRF52840 Express <https://www.adafruit.com/product/4062>
   Adafruit Circuit Playground Bluefruit <https://www.adafruit.com/product/4333>

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

"""
try:
    from typing import Optional, Tuple
    import _bleio
    from circuitpython_typing import ReadableBuffer
except ImportError:
    pass


import time
import struct
from micropython import const
from adafruit_ble import BLERadio
from adafruit_ble.advertising import Advertisement, LazyObjectField
from adafruit_ble.advertising.standard import ManufacturerData
from adafruit_ble.advertising.adafruit import (
    MANUFACTURING_DATA_ADT,
    ADAFRUIT_COMPANY_ID,
)


__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_BLE_Radio.git"


#: Maximum length of a message (in bytes).
MAX_LENGTH = 248

#: Amount of time to advertise a message (in seconds).
AD_DURATION = 0.5

_RADIO_DATA_ID = const(0x0001)  # TODO: check this isn't already taken.


class _RadioAdvertisement(Advertisement):
    """Broadcast arbitrary bytes as a radio message."""

    match_prefixes = (struct.pack("<BH", 0xFF, ADAFRUIT_COMPANY_ID),)
    manufacturer_data = LazyObjectField(
        ManufacturerData,
        "manufacturer_data",
        advertising_data_type=MANUFACTURING_DATA_ADT,
        company_id=ADAFRUIT_COMPANY_ID,
        key_encoding="<H",
    )

    @classmethod
    def matches(cls, entry: _bleio.ScanEntry) -> bool:
        """Checks for ID matches"""
        if len(entry.advertisement_bytes) < 6:
            return False
        # Check the key position within the manufacturer data. We already know
        # prefix matches so we don't need to check it twice.
        return (
            struct.unpack_from("<H", entry.advertisement_bytes, 5)[0] == _RADIO_DATA_ID
        )

    @property
    def msg(self) -> ReadableBuffer:
        """Raw radio data"""
        if _RADIO_DATA_ID not in self.manufacturer_data.data:
            return b""
        return self.manufacturer_data.data[_RADIO_DATA_ID]

    @msg.setter
    def msg(self, value: ReadableBuffer) -> None:
        self.manufacturer_data.data[_RADIO_DATA_ID] = value


class Radio:
    """
    Represents a connection through which one can send or receive strings
    and bytes. The radio can be tuned to a specific channel upon initialisation
    or via the `configure` method.
    """

    def __init__(self, **args):
        """
        Takes the same configuration arguments as the `configure` method.
        """
        # For BLE related operations.
        self.ble = BLERadio()
        # The uid for outgoing message. Incremented by one on each send, up to
        # 255 when it's reset to 0.
        self.uid = 0
        # Contains timestamped message metadata to mitigate report of
        # receiving of duplicate messages within AD_DURATION time frame.
        self.msg_pool = set()
        # Handle user related configuration.
        self.configure(**args)

    def configure(self, channel: int = 42) -> None:
        """
        Set configuration values for the radio.

        :param int channel: The channel (0-255) the radio is listening /
            broadcasting on.
        """
        if -1 < channel < 256:
            self._channel = channel
        else:
            raise ValueError("Channel must be in range 0-255")

    def send(self, message: str) -> None:
        """
        Send a message string on the channel to which the radio is
        broadcasting.

        :param str message: The message string to broadcast.
        """
        return self.send_bytes(message.encode("utf-8"))

    def send_bytes(self, message: bytes) -> None:
        """
        Send bytes on the channel to which the radio is broadcasting.

        :param bytes message: The bytes to broadcast.
        """
        # Ensure length of message.
        if len(message) > MAX_LENGTH:
            raise ValueError("Message too long (max length = {})".format(MAX_LENGTH))
        advertisement = _RadioAdvertisement()
        # Concatenate the bytes that make up the advertised message.
        advertisement.msg = struct.pack("<BB", self._channel, self.uid) + message

        self.uid = (self.uid + 1) % 255
        # Advertise (block) for AD_DURATION period of time.
        self.ble.start_advertising(advertisement)
        time.sleep(AD_DURATION)
        self.ble.stop_advertising()

    def receive(self, timeout: float = 1.0) -> str:
        """
        Returns a message received on the channel on which the radio is
        listening.

        :param float timeout: The length of time (in seconds) the radio listens for a broadcast

        :return: A string representation of the received message, or else None.
        """
        msg = self.receive_full(timeout=timeout)
        if msg:
            return msg[0].decode("utf-8").replace("\x00", "")
        return None

    def receive_full(
        self, timeout: float = 1.0
    ) -> Optional[Tuple[ReadableBuffer, int, float]]:
        """
        Returns a tuple containing three values representing a message received
        on the channel on which the radio is listening. If no message was
        received then `None` is returned.

        The three values in the tuple represent:

        * the bytes received.
        * the RSSI (signal strength: 0 = max, -255 = min).
        * a microsecond timestamp: the value returned by time.monotonic() when
          the message was received.

        :param float timeout: The length of time (in seconds) the radio listens for a broadcast

        :return: A tuple representation of the received message, or else None.

        """
        try:
            for entry in self.ble.start_scan(
                _RadioAdvertisement, minimum_rssi=-255, timeout=timeout, extended=True
            ):
                # Extract channel and unique message ID bytes.
                chan, uid = struct.unpack("<BB", entry.msg[:2])
                if chan == self._channel:
                    now = time.monotonic()
                    addr = entry.address.address_bytes
                    # Ensure this message isn't a duplicate. Message metadata
                    # is a tuple of (now, chan, uid, addr), to (mostly)
                    # uniquely identify a specific message in a certain time
                    # window.
                    expired_metadata = set()
                    duplicate = False
                    for msg_metadata in self.msg_pool:
                        if msg_metadata[0] < now - AD_DURATION:
                            # Ignore expired entries and mark for removal.
                            expired_metadata.add(msg_metadata)
                        elif (chan, uid, addr) == msg_metadata[1:]:
                            # Ignore matched messages to avoid duplication.
                            duplicate = True
                    # Remove expired entries.
                    self.msg_pool = self.msg_pool - expired_metadata
                    if not duplicate:
                        # Add new message's metadata to the msg_pool and
                        # return it as a result.
                        self.msg_pool.add((now, chan, uid, addr))
                        msg = entry.msg[2:]
                        return (msg, entry.rssi, now)
        finally:
            self.ble.stop_scan()
        return None



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
#
# SPDX-License-Identifier: Unlicense

"""
This file ensures everything is in place to run PyTest based unit tests against
the adafruit_radio module. It works by using Python's mock library to add
MagicMock objects to sys.modules for the modules which are not available to
standard Python because they're CircuitPython only modules.

Such mocking happens as soon as this conftest.py file is imported (so the
mocked modules exist in sys.modules before the module to be tested is
imported), and immediately before each test function is evaluated (so changes
to state remain isolated between tests).
"""
import sys
from unittest.mock import MagicMock

# Add fully qualified namespace paths to things that are imported, but which
# should be mocked away. For instance, modules which are available in
# CircuitPython but not standard Python.
MOCK_MODULES = [
    "adafruit_ble.BLERadio",
    "adafruit_ble.advertising.adafruit.AdafruitRadio",
]


def mock_imported_modules():
    """
    Mocks away the modules named in MOCK_MODULES, so the module under test
    can be imported with modules which may not be available.
    """
    module_paths = set()
    for mock in MOCK_MODULES:
        namespaces = mock.split(".")
        namespace = []
        for n in namespaces:
            namespace.append(n)
            module_paths.add(".".join(namespace))
    for m_path in module_paths:
        sys.modules[m_path] = MagicMock()


def pytest_runtest_setup():
    """
    Called immediately before any test function is called.

    Recreates afresh the mocked away modules so state between tests remains
    isolated.
    """
    mock_imported_modules()


# Initial mocking needed to stop ImportError when importing module under test.
mock_imported_modules()



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
#
# SPDX-License-Identifier: Unlicense

"""
Simple unit tests for the adafruit_ble_radio module. Uses experimental mocking
found in the testconf.py file. See comments therein for explanation of how it
works.
"""
import struct
import time
from unittest import mock
import pytest
import adafruit_ble_radio


@pytest.fixture
def radio():
    """
    A fixture to recreate a new Radio instance for each test that needs it.
    """
    return adafruit_ble_radio.Radio()


def test_radio_init_default():
    """
    Ensure a Radio object is initialised in the expected way:

    * It has a BLERadio instance.
    * The self.uid counter is set to 0.
    * The self.msg_pool is initialised as an empty set.
    * The channel is set to the default 42.
    """
    r = adafruit_ble_radio.Radio()
    assert r.ble == adafruit_ble_radio.BLERadio()
    assert r.uid == 0
    assert r.msg_pool == set()
    assert r._channel == 42  # pylint: disable=protected-access


def test_radio_init_channel():
    """
    If a channel argument is passed to initialisation, this is correctly set.
    """
    r = adafruit_ble_radio.Radio(channel=7)
    assert r._channel == 7  # pylint: disable=protected-access


def test_radio_configure_channel(radio_obj):
    """
    If a valid channel argument is passed to the configure method, the Radio
    instance's channel is updated to reflect this.
    """
    assert radio_obj._channel == 42  # pylint: disable=protected-access
    radio_obj.configure(channel=7)
    assert radio_obj._channel == 7  # pylint: disable=protected-access


def test_radio_configure_channel_out_of_bounds(
    radio_obj,
):  # pylint: disable=invalid-name
    """
    If a channel not in the range 0-255 is passed into the configure method,
    then a ValueError exception is raised.
    """
    with pytest.raises(ValueError):
        radio_obj.configure(channel=-1)
    with pytest.raises(ValueError):
        radio_obj.configure(channel=256)
    # Add just-in-bounds checks too.
    radio_obj.configure(channel=0)
    assert radio_obj._channel == 0  # pylint: disable=protected-access
    radio_obj.configure(channel=255)
    assert radio_obj._channel == 255  # pylint: disable=protected-access


def test_radio_send(radio_obj):
    """
    The send method merely encodes to bytes and calls send_bytes.
    """
    radio_obj.send_bytes = mock.MagicMock()
    msg = "Testing 1, 2, 3..."
    radio_obj.send(msg)
    radio_obj.send_bytes.assert_called_once_with(msg.encode("utf-8"))


def test_radio_send_bytes_too_long(radio_obj):
    """
    A ValueError is raised if the message to be sent is too long (defined by
    MAX_LENGTH).
    """
    msg = bytes(adafruit_ble_radio.MAX_LENGTH + 1)
    with pytest.raises(ValueError):
        radio_obj.send_bytes(msg)


def test_radio_send_bytes(radio_obj):
    """
    Ensure the expected message is set on an instance of AdafruitRadio, and
    broadcast for AD_DURATION period of time.
    """
    radio_obj.uid = 255  # set up for cycle back to 0.
    msg = b"Hello"
    with mock.patch("adafruit_ble_radio.time.sleep") as mock_sleep:
        radio_obj.send_bytes(msg)
        mock_sleep.assert_called_once_with(adafruit_ble_radio.AD_DURATION)
    spy_advertisement = (
        adafruit_ble_radio._RadioAdvertisement()  # pylint: disable=protected-access
    )  # pylint: disable=protected-access
    chan = struct.pack("<B", radio_obj._channel)  # pylint: disable=protected-access
    uid = struct.pack("<B", 255)
    assert spy_advertisement.msg == chan + uid + msg
    radio_obj.ble.start_advertising.assert_called_once_with(spy_advertisement)
    radio_obj.ble.stop_advertising.assert_called_once_with()
    assert radio_obj.uid == 0


def test_radio_receive_no_message(radio_obj):
    """
    If no message is received from the receive_bytes method, then None is
    returned.
    """
    radio_obj.receive_full = mock.MagicMock(return_value=None)
    assert radio_obj.receive() is None
    radio_obj.receive_full.assert_called_once_with()


def test_radio_receive(radio_obj):
    """
    If bytes are received from the receive_bytes method, these are decoded
    using utf-8 and returned as a string with null characters stripped from the
    end.
    """
    # Return value contains message bytes, RSSI (signal strength), timestamp.
    msg = b"testing 1, 2, 3\x00\x00\x00\x00\x00\x00"
    radio_obj.receive_full = mock.MagicMock(return_value=(msg, -20, 1.2))
    assert radio_obj.receive() == "testing 1, 2, 3"


def test_radio_receive_full_no_messages(radio_obj):  # pylint: disable=invalid-name
    """
    If no messages are detected by receive_full then it returns None.
    """
    radio_obj.ble.start_scan.return_value = []
    assert radio_obj.receive_full() is None
    radio_obj.ble.start_scan.assert_called_once_with(
        adafruit_ble_radio._RadioAdvertisement,  # pylint: disable=protected-access
        minimum_rssi=-255,
        timeout=1,
        extended=True,
    )
    radio_obj.ble.stop_scan.assert_called_once_with()


def test_radio_receive_full_duplicate_message(
    radio_obj,
):  # pylint: disable=invalid-name
    """
    If a duplicate message is detected, then receive_full returns None
    (indicating no *new* messages received).
    """
    mock_entry = mock.MagicMock()
    mock_entry.msg = b"*\x00Hello"
    mock_entry.address.address_bytes = b"addr"
    mock_entry.rssi = -40
    radio_obj.ble.start_scan.return_value = [mock_entry]
    radio_obj.msg_pool.add((time.monotonic(), 42, 0, b"addr"))
    assert radio_obj.receive_full() is None


def test_radio_receive_full_and_remove_expired_message_metadata(
    radio_obj,
):  # pylint: disable=invalid-name
    """
    Return the non-duplicate message.

    Ensure that expired message metadata (used to detect duplicate messages
    within a short time window) is purged from the self.msg_pool cache.

    Ensure the metadata from the new message is now in the self.msg_pool cache.
    """
    mock_entry = mock.MagicMock()
    mock_entry.msg = b"*\x01Hello"
    mock_entry.address.address_bytes = b"adr2"
    mock_entry.rssi = -40
    radio_obj.ble.start_scan.return_value = [mock_entry]
    radio_obj.msg_pool.add(
        (time.monotonic() - adafruit_ble_radio.AD_DURATION - 1, 42, 0, b"addr")
    )
    result = radio_obj.receive_full()
    assert result[0] == b"Hello"
    assert result[1] == -40
    assert len(radio_obj.msg_pool) == 1
    metadata = radio_obj.msg_pool.pop()
    assert metadata[1] == 42
    assert metadata[2] == 1
    assert metadata[3] == b"adr2"



# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT

"""
This demo uses the adafruit_radio module to send and receive messages.
Devices are switched between broadcast and scanning using the slide switch.
The buttons change the message to be sent.
"""
import digitalio
import board
from adafruit_ble_radio import Radio


slide_switch = digitalio.DigitalInOut(board.SLIDE_SWITCH)
slide_switch.pull = digitalio.Pull.UP
button_a = digitalio.DigitalInOut(board.BUTTON_A)
button_a.pull = digitalio.Pull.DOWN
button_b = digitalio.DigitalInOut(board.BUTTON_B)
button_b.pull = digitalio.Pull.DOWN

led = digitalio.DigitalInOut(board.D13)
led.switch_to_output()

msg = [
    "hello",
    "hi",
    "foo",
    "bar",
    "baz",
]

i = 0
r = Radio()

while True:
    if slide_switch.value:
        print("Sending messages...")
        while slide_switch.value:
            last_i = i
            if button_a.value:
                i += 1
            if button_b.value:
                i -= 1
            i %= len(msg)
            m = msg[i]
            print("Sending {}".format(m))
            r.send(m)
            # Alternative
            # r.send_bytes(b"Arbitrary bytes")
    else:
        print("Scanning for messages...")
        while not slide_switch.value:
            m = r.receive_full()
            if m:
                print("Received message: {}".format(m))
            # Alternative
            # m = r.receive()
            # if m:
            #    print(m)
